/**************************************************************************************

 Description:
                           
    Asm45:  Symbolic assembler/disassembler for HP9845 object code
     
    For the 45 assembler language, please refer to HP's "Assembly Development
    ROM Manual for the HP9845". Mostly everything in that manual will be
    supported by this implementation of a PC-to-System45 cross assembler, with
    some specialties mentioned below.
    
    See http://www.hp9845.net/projects/assembler for additional information.
 
 System requirements:
 
    Runs as console application under any WIN32 environment (Win9x up to Vista).
 
 Getting started:
 
    Start a console window (command.exe for Windows 9x/ME and cmd.exe for
    Windows NT/2000/XP/Vista).
    
    From the console, executing Asm45 with
    
        asm45.exe <source_file>
        
    will run the assembler on the source file and produce the appropriate object
    code in a newly created file out.obj. <source_file> must be a valid system
    45 assembler program (see the HP9845 Assembly Development ROM Manual for
    System 45 Assembler syntax).
    
    Executing Asm45 with
    
        asm45.exe -d <object_file> > <output_file>
        
    will produce a disassembly of <object_file> into <output_file>.
    
    Execute with -h option for further usage information.
 
 Restrictions:
 
    Up to 10000 symbols, 10000 labels and 100 literal pools are supported.
    Line length is 160 characters maximum, although 80 characters are
    recommended for readability.

 Compatibility notes:
 
 1) 'Clear/set' bit for test/alter/branch group in hold state is not defined
    and will be defaulted to 'clear' (bit 6 = zero). Doesn't matter for the
    processor, however may produce slightly different object code than with HP's
    assemblers.
    
 2) BCD arithmetics are different for some special cases. IEEE754 specifies
    binary base exponents, the system 45 processors work with exponents to the
    base 10. The exponent range of the system 45 is about one-third smaller.
    In addition, the HP assembler limits input and output of exponents from
    -99 up to +99. IEEE754 defines special numbers (e.g. infinite) which have
    no equivalence in the HP floating point format.
    
    Floating point arithmetics (expressions with floating point numbers) are
    supported, but are not by the original HP assembler.
    
 3) All 9845 assembler code will be accepted, however source listing control
    and conditional compilation (IFA, IFB, etc.) are not yet implemented
    and therefore have no effect. Preceding line numbers or addresses
    in assembly code lines will be intentionally ignored for maximum
    compatibility.
    
 4) One pseudo opcode (ORG) has been borrowed from ASMA, HP's original '45
    cross assembler. This pseudo opcode may be used to set/reset the current
    code address counter and may be used for assembling separate portions
    of code or for including external references (e.g. base page definitions).
    ORG instructions may be used anywhere in the code, however it is strongly
    recommended to use one single instance in each module immediately before
    the NAM instruction. If ORG instructions are ommitted, modules start at
    default location (=0) and are addressed in ascending order, each starting
    after the end of the previous one.
    
 5) Semicolon and brackets are supported as additional means for comments.
    The semicolon is equivalent to the exclamation mark and informs the
    assembler to ignore the rest of the line. Brackets may be used for
    in-line-comments (especially useful for information automatically
    inserted by the disassembler). Finally, a line starting with an asterisk
    ('*') as very first character are completely treated as comment (for
    compatibility towards ASMA).
    
 6) Parsing and syntax error checking are somewhat relaxed in order to process
    as many '45 assembler styles as possible. Case will be normalized automati-
    cally (especially for symbol names), labels must be completed with a colon,
    and opcodes must be three-character-uppercase and separated from operand by
    whitespace. Anything else is free or will just be ignored.
    
 7) Support for hexadecimal and binary numbers has been added. Hexadecimal
    numbers start with a dollar sign (like $07ff) and binary numbers end with
    an uppercase 'L'(e.g. 00110110L). Support for octal representation is just
    the same as in HP's assembler (octal numbers end with an uppercase 'B').
    All addresses may be output either as hexadecimal (default) or octal (use
    the -o command line switch) numbers.
    
 8) String concatenation with the '+' operator is allowed (e.g. "123"+"ABC" =
    "123ABC").
 
 9) For convenience, additional predefined symbols give direct access to the
    individual words of the BCD accumulators, base page temporary area or
    utility temporary storage area.
    
 10) The assembler provides a built-in disassembler plus some extra tools
     for building cross-reference tables, string tables, symbol tables or
     memory dumps. The cross-refrence table identifies constants from variables
     just through the type of access: locations, which are only referenced by
     read instructions, are assumed as constants. Locations which are
     written to are considerd as variables. Assembler and disassembler are
     reversable, i.e. the assembler may be run on the output of the disassem-
     bler, reproducing the original machine code with the exception of (1)
     and (14).

     Another situation where assembly and disassembly may differ is non-
     arithmetic left/right rotation (RAR/RAL and RBR/RBL) where right and
     left shift are actually implemented through the same object code, which
     in turn makes it impossible for the disassembler to reconstruct the
     original assembler instruction. However, machine code (and computation
     results) will be the same.
     
 11) Literal pools will be filled straight-forward, i.e. in the order of use.
     Boundaries will be checked, and the best-fit literal pool will be
     chosen, however there is no kind of sophisticated intra-pool sorting
     optimization. Planning literal spaces is up the the programmer.
     
 12) There is no difference between the EQU and the SET pseudo opcodes, both
     may be used for re-defining symbols at any time. The restriction of the
     HP assembler has intentionally been relaxed. Same for the definition of
     'complex' literals. Both floating point numbers and strings may be
     defined as literals with the '=' operator. Although there is no serious
     need for '==' operator, it is still supported for the sake of
     compatibility.
     
 13) Automatic indirection is fully supported and may be switched on and off
     just like in HP's assembler via ION and IOF pseudo instructions. In case
     the assembler makes use of this feature, a corresponding comment is
     generated in the produced listing indicating the change.
     
 14) In order to generate relocatable code, the assembler automatically
     assigns addresses of type absolute within octal 0B and 777B or between
     177000B and 177777B as 'base page' reference. As such, those addresses
     are normally 'external' to the program code, since they are hold
     e.g. in the system roms and are nor part of the program itself.
     
     However, some applications require absolute references to labels within
     the base page area without base page addressing. This is especially
     true for system ROMs, since they cover the lower base page area for
     storing system constants. Normally label references within the program
     code are always of type relocatable. For system ROMs, references to
     labels located in the base page area should be of type absolute since
     they have to be referenced from the whole address space.
     
     All label references within the pseudo instructions BON and BOF
     (Base page ON, Base page OFF) are treated as absolute addresses, if
     they are located within the base page range (outside the base page
     area both BON and BOF have no effect).
     
     The default behavior may be re-defined using the '-r' command line
     option (until the first occurance of BON or BOF, which have higher
     precedence).
     
     For the disassembler, the thing is similar. Unless the -r option
     isn't used, all base page references will be handled as external
     references (i.e. will not produce a label unless -ppu or -lpu options
     are used). If you want to disassemble firmware including references to
     the base page, use the -ppu or -lpu option for a proper disassembly,
     depending on whether the code is intended to run on the PPU or on the
     LPU.
     
     Another way to force base page referencing is to define symbols with the
     EQU statement before the program code, like
     
        <Symbol>:  EQU A+<location>
     
     which will automatically be translated as base page reference due to
     the offset to the A register.
     
 15) If programming binary programs or option ROM, in general base page
     addressing is used to access system constants and/or system variables.
     If using the -lpu or -ppu options with disassembling, all base page
     accesses are named with the appropriate symbols as labels, and also
     a short comment on the semantics of the location is printed.
     
     Use -lpu for code intended to run on the LPU (e.g. binary programs, or
     LPU option ROMs) and -ppu for code intended to run on the PPU. All symbols
     refer to the HP9845B/C (Galleon/Odyssey) systems, you should not use the
     -lpu or -ppu option for HP9845A (Qwert) or HP9835 (Raven) code.
     
     The options -r, -lpu and -ppu should be used exclusively, do not mix them
     together.
    
 16) Modules are handled separately. Each module has its own symbol space,
     literal space and so on. Object code for modules will be generated into
     seperate files, each using the module name as filename. By specifying a
     <output file> via the command line, all modules will be generated into
     that single file (in ascending order).

 Application notes:

  1) Producing relocatable code: there is a strong difference between designing
     relocatable source code and relocatable object code. Relocatable source
     code means code which may be re-assambled with another stating address,
     relocatable object code means code, which may be loaded anywhere into
     memory without re-assembly. Relocatable source code may be easily achieved
     by just restricting all absolute references to the base page area.
     
     Relocatable objet code is somewhat more demanding, since any non-relative
     reference within the code has to be avoided (again with the exception of
     the base page area). This includes the total avoidence of literals. Any
     references to absolute addresses within the code (e.g. for passing para-
     meters to external routines) have to be calculated with respect to the
     program counter (R02 or "P") during run-time.
 
 History:

	Version 1.1:
		- added base page labels & comments 

    Version  1.0 beta:
    	- added base page ROM constants

    Version  1.0 alpha (fully functional, but not thoroughly tested)

 Things to do:
 
    - Directive to include external file content (e.g. base page definitions)
    - Escape sequences in strings for control characters & double quotes
    - Support for conditional assembly & comparison operators
    - Support for external symbol defintions & macros (linker)
    - Code library, ROM entry points, statistics...
    - Probably some more bugfixing
    
 Building:

    The Asm45 executable was built with the free MinGW GNU C compiler
    version 3.4.5 and should compile with any GNU C compiler. An appropriate
    makefile is included in the source package.

 License information:

    Copyright (c) 2009 Ansgar Kueckes

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
   
***************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "../include/asm45.h"

#define VERSIONSTRING	"1.1 (2009/10/08)"

/* globals */

unsigned int *rom_lower_base_page;

Op_code *op_table;
int op_table_size;
int num_predefined = 0;		// number of predefined symbols, will be set in main() section

Symbol symbol_table[MAX_SYMBOLS];

Literalspace literalspace_table[MAX_LITERALSPACES];

int symbols_size = 0;
int literalspaces_size = 0;

int longest_symbol_name = 0;

Label internals[MAX_LABEL];
Label externals[MAX_LABEL];

int internals_size = 0;
int externals_size = 0;

int num_error_messages;

char **rom_labels;
char **ram_labels;
char **rom_comments;
char **ram_comments;
int rom_labels_access[512];
int ram_labels_access[512];

/* switches */
int verbose = 0;
int octal = 0;
int ascii = 0;
int comments = 0;
int clocks = 0;
int offset = 0;
int start = 0;
int end = 0;
int xref = 0;
int dump = 0;
int disassembly = 0;
int strings = 0;
int listing = 0;
int symbols = 0;
int numbers = 0;
int base_page_mode = 0;
int map_basepage = 0;
int list_basepage_symbols = 0;
int isource = 0;

unsigned int filesize = 0;

/* == general routines ================================================================ */

void show_short_license(char *program, int year)
{
	printf("%s Copyright (c) %d Ansgar Kueckes\n", program, year);
    printf("This program comes with ABSOLUTELY NO WARRANTY.\n");
    printf("This is free software, and you are welcome to redistribute it\n");
    printf("under certain conditions. Type '%s -license' for details.\n", program);
}

void show_full_license()
{
	FILE *license;
	char line[256];
	int count = 0;
	
	license = fopen("COPYING", "r");
	if (!license) license = fopen("../COPYING", "r");
	if (!license) license = fopen("../../COPYING", "r");
	if (!license) {
		printf("\tThis program is free software: you can redistribute it and/or modify\n");
		printf("\tit under the terms of the GNU General Public License as published by\n");
		printf("\tthe Free Software Foundation, either version 3 of the License, or\n");
		printf("\tany later version.\n\n");
		printf("\tThis program is distributed in the hope that it will be useful,\n");
		printf("\tbut WITHOUT ANY WARRANTY; without even the implied warranty of\n");
		printf("\tMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
		printf("\tGNU General Public License for more details.\n\n");
		printf("\tYou should have received a copy of the GNU General Public License\n");
		printf("\talong with this program. If not, see <http://www.gnu.org/licenses/>.\n");		
	}
	else {
		while(fgets(line, 256, license)) {
			printf("%s", line);
			if (count++ > 23) {
				printf("\nType <enter> for more or q <enter> to quit...");
				gets(line);
				if ((line[0] == 'q') || (line[0] == 'Q')) return;
				count = 0;
			}
		}
		fclose(license);
	}
}

void ordered_insert_string(Stringlist **plist, char *str, int do_sort)
{
	Stringlist *indx, *node, *last;

	if (!str) return;
	if (!plist) return;

	/* create node */
	node = (Stringlist *)malloc(sizeof(Stringlist));
	if (!node) {
		printf("Out of memory error\n");
		exit(-1);
	}
	node->string = (char *)malloc(strlen(str)+1);
	if (!node->string) {
		printf("Out of memory error\n");
		exit(-1);
	}
	node->next = NULL;
	strcpy(node->string, str);
	if (!*plist) *plist = node;
	else {
  		/* find position */
  		last = NULL;
  		for (indx=*plist; indx; indx=(Stringlist *)indx->next) {
  			if (do_sort && strcmp(indx->string, node->string) > 0) break;
  			last = indx;
  		}
		/* insert/append */
		node->next = (Stringlist *)indx;
		if (last) last->next = node;
		else *plist = node;
	}
}

void print_usage(char *command)
{
	fprintf(stderr, "Usage: %s [-acdlmnostvx][-i <address>][-w <start>,<end>] <input file> [<output file>]\n", command);
	fprintf(stderr, "        <input file>     HP 9845 input file (assembler source or object binary)\n");
	fprintf(stderr, "        <output file>    HP 9845 output file (object binary or disassembly)\n");
	fprintf(stderr, "        -i <address>     initial memory address (default: 0)\n");
	fprintf(stderr, "        -o               octal data presentation (default: hex)\n");
	fprintf(stderr, "        -d               produce disassembly (default: assemble)\n");
	fprintf(stderr, "        -n               activate line numbering\n");
	fprintf(stderr, "        -r               system ROM mode (treat base page references as local)\n");
	fprintf(stderr, "        -v               verbose mode\n");
	fprintf(stderr, "        -version         show version info\n");
	fprintf(stderr, "        -license         show license info\n");
	fprintf(stderr, "\n     assembler options:\n");
	fprintf(stderr, "        -l               produce listing\n");
	fprintf(stderr, "        -s               build symbol table\n");
	fprintf(stderr, "\n     disassembler options:\n");
	fprintf(stderr, "        -isource         use Assy Exec & Dev ROM syntax\n");
	fprintf(stderr, "        -w <start>,<end> memory window for disassembly\n");
	fprintf(stderr, "        -lpu             map LPU ROM constants space into base page range\n");
	fprintf(stderr, "        -ppu             map PPU ROM constants space into base page range\n");
	fprintf(stderr, "        -b               include used base page symbols (with -lpu or -ppu)\n");
	fprintf(stderr, "        -a               include ASCII representation\n");
	fprintf(stderr, "        -t               include timing information (clock cycles)\n");
	fprintf(stderr, "        -c               include opcode short description\n");
	fprintf(stderr, "        -x               build cross reference\n");
	fprintf(stderr, "        -s               build string table\n");
	fprintf(stderr, "        -m               output memory dump\n");
	return;
}

void print_error(int line, int err_no)
{
	if (err_no<0) err_no = -err_no;
	if (err_no<=num_error_messages)
		fprintf(stderr, "Line %i: Error: %s.\n", line, error_messages[err_no]);
	return;
}

void printf_string(char *str, char *format, ...)
{
   int len;
   va_list argp;
   
   if (!str) return;
   len = strlen(str);
   va_start(argp, format);
   vsprintf(str+len, format, argp);	// append
   va_end(argp);
}

int get_cmd_type(unsigned int cmd)
/*
	1 = constant (read-only access)
	2 = variable (read/write access)
	3 = constant vector (indirect read-only access)
	4 = variable vector (indirect read/write access)
	5 = subroutine
	6 = jump point
	7 = subroutine vector (indirect JSM)
	8 = jump vector (indirect JMP)
	9 = branch (test/alter/skip)
	0 = other (register etc.)
*/
{
    unsigned int tmp;
    
    tmp = (cmd >> 10) & 0x3F;
    
    if (((tmp&0x3C)<0xC) ||			// LDx, CPx, ADx
        ((tmp&0x3E)==0x14) ||		// AND
        ((tmp&0x3E)==0x18))			// IOR
        return 1;
    else if (((tmp&0x3C)==0xC) ||	// STx
	     ((tmp&0x3E)==0x12) ||		// ISZ
	     ((tmp&0x3E)==0x13))		// DSZ
	return 2;
    else if (((tmp&0x1C)<0xC) ||	// LDx, CPx, ADx indirect
             ((tmp&0x1E)==0x14) ||	// AND indirect
             ((tmp&0x1E)==0x18))	// IOR indirect
        return 3;
    else if (((tmp&0x1C)==0xC) ||	// STx indirect
	     ((tmp&0x1E)==0x12) ||		// ISZ indirect
	     ((tmp&0x1E)==0x13))		// DSZ indirect
	return 4;
    else if ((tmp&0x3E)==0x10)		// JSM
    	return 5;
    else if ((tmp&0x3E)==0x1A)		// JMP
    	return 6;
    else if ((tmp&0x3E)==0x30)		// JSM indirect
    	return 7;
    else if ((tmp&0x3E)==0x3A)		// JMP indirect
    	return 8;
    else if ((tmp&0x1D)==0x1D)		// test/alter/skip
    	return 9;
    else return 0;
}

void print_xref(Label *list, int size, int type, int isexternal)
{
    int i, count;
    Node *node;
    
    for (i=0; i<size; i++) {
    	count = list[i].reftype[type];
    	if ((type == 1) && (list[i].reftype[2] > 0)) count = 0;		// Read access
    	if ((type == 2) && (count > 0)) count += list[i].reftype[1];	// Write access
    	if ((type == 3) && (list[i].reftype[4] > 0)) count = 0;		// Indirect read access
    	if ((type == 4) && (count > 0)) count += list[i].reftype[3];	// Indirect write access
	if (count > 0) {
	  if (isexternal) {
	  	if (map_basepage && (list[i].address >= 0x20) && (list[i].address < 0x200)) {
	     if (octal) printf("%06o (=%06o) [%d/%d]\t",
	  		list[i].address, rom_lower_base_page[list[i].address], count, list[i].refcount);
	     else printf("%04x (=%04x) [%d/%d]\t",
	  		list[i].address, rom_lower_base_page[list[i].address], count, list[i].refcount);
	  	}
	  	else {
	     if (octal) printf("%06o [%d/%d]\t",
	  		list[i].address, count, list[i].refcount);
	     else printf("%04x [%d/%d]\t",
	  		list[i].address, count, list[i].refcount);
	  	}
	  }
	  else {
	     if (octal) printf("%06o (=%06o) Lbl_%04d [%d/%d]\t",
	  	list[i].address, list[i].value, i+1, count, list[i].refcount);
	     else printf("%04x (=%04x) Lbl_%04d [%d/%d]\t",
	  	list[i].address, list[i].value, i+1, count, list[i].refcount);
	  }
	  if (count > list[i].reftype[type]) printf("W: ");	// Variable
	  node = list[i].ref[type];
	  while (node) {
	      if (octal) printf("%06o ", node->address);
	      else printf("%04x ", node->address);
	      node = node->next;
	  }
	  if (count > list[i].reftype[type]) {			// Special handling for variables
	  	printf("R: ");
		node = list[i].ref[type-1];
		while (node) {
		   if (octal) printf("%06o ", node->address);
		   else printf("%04x ", node->address);
		   node = node->next;
		}
	  }
	  printf("\n");
	}
    }
}


/* == main =========================================================================== */

int main(int argc, char *argv[])
{
  FILE *ifp, *ofp = 0;
  char *outfile = 0;
  char *infile = 0;
  char *range = 0;
  char *pos;
  unsigned char c;
  unsigned char buffer[1024];
  unsigned char str[1024];
  unsigned char space[] = "         ";
  unsigned int cmd = 0;
  unsigned int address = 0;
  int rec_size = 256;
  int seek_start = 0;
  int window = 0;
  int first = 0;
  int i, j, count, label, line;
  int input_line;
  size_t bytes_read;
  Node *node,*tmp;
  P_symbol *psymbol;
  Symbol tmp_symbol;

  if (argc < 2) { print_usage(argv[0]); exit(0); }

  for ( count = 1; count < argc ; count++)
  {
	if (!strcmp(argv[count], "-d") && (count < argc))
		{ disassembly = 1; continue; }

	if (!strcmp(argv[count], "-t") && (count < argc))
		{ clocks = 1; continue; }

	if (!strcmp(argv[count], "-ppu") && (count < argc))
		{
			map_basepage = 1;
			rom_lower_base_page = ppu_rom_lower_base_page;
			rom_labels = ppu_rom_labels;
			ram_labels = ppu_ram_labels;
			rom_comments = ppu_rom_comments;
			ram_comments = ppu_ram_comments;
			continue;
		}

	if (!strcmp(argv[count], "-lpu") && (count < argc))
		{
			map_basepage = 1;
			rom_lower_base_page = lpu_rom_lower_base_page;
			rom_labels = lpu_rom_labels;
			ram_labels = lpu_ram_labels;
			rom_comments = lpu_rom_comments;
			ram_comments = lpu_ram_comments;
			continue;
		}

	if (!strcmp(argv[count], "-b"))
		{
			list_basepage_symbols = 1;
			continue;
		}

	if (!strcmp(argv[count], "-c") && (count < argc))
		{ comments = 1; continue; }

	if (!strcmp(argv[count], "-r") && (count < argc))
		{ base_page_mode = 1; continue; }

	if (!strcmp(argv[count], "-v") && (count < argc))
		{ verbose = 1; continue; }

	if (!strcmp(argv[count], "-o") && (count < argc))
		{ octal = 1; continue; }

	if (!strcmp(argv[count], "-x") && (count < argc))
		{ xref = 1; continue; }

	if (!strcmp(argv[count], "-m") && (count < argc))
		{ dump = 1; continue; }

	if (!strcmp(argv[count], "-n") && (count < argc))
		{ numbers = 1; continue; }

	if (!strcmp(argv[count], "-s") && (count < argc))
		{ strings = 1; symbols = 1; continue; }

	if (!strcmp(argv[count], "-a") && (count < argc))
		{ ascii = 1; continue; }
		
	if (!strcmp(argv[count], "-l") && (count < argc))
		{ listing = 1; continue; }
		
	if (!strcmp(argv[count], "-isource") && (count < argc))
		{ isource = 1; continue; }

	if (window == -1)
		{ range = argv[count]; window = 1; continue; };

	if (!strcmp(argv[count], "-w") && (count < argc))
		{ window = -1; continue; }

	if (first == -1)
		{ offset = strtoul(argv[count],0,0) & 0xFFFF; first = 1; continue; };

	if (!strcmp(argv[count], "-i") && (count < argc))
		{ first = -1; continue; }
		
	if (!strcmp(argv[count], "-version"))
		{ printf("%s\n", VERSIONSTRING); show_short_license("Asm45", 2009); exit(0); }

	if (!strcmp(argv[count], "-license"))
		{ show_full_license(); exit(0); }
		
	if (!strncmp(argv[count], "-", 1) && (count < argc))
		 { print_usage(argv[0]); exit(0); }
		 
	if (!infile) infile = argv[count];
	else if (!outfile) outfile = argv[count];
  }
  if (!infile) { print_usage(argv[0]); exit(0); }
  
  ifp = fopen( infile, "rb");
  if (!ifp) { fprintf(stderr, "Could not open file \"%s\".\n", infile); exit(-1); }

  if (list_basepage_symbols) {
  	for (i=0; i<512; i++) {
  		ram_labels_access[i] = 0;
  		rom_labels_access[i] = 0;
  	}
  }

  while (fgetc(ifp) != EOF) filesize++;
  filesize = (filesize + 1) / 2;
  
  for (num_error_messages=0; error_messages[num_error_messages]; num_error_messages++);
  
  //for (i=0; i<10; i++) space[i] = ' ';
  //space[10] = 0;
  
  /* get range */
  start = offset;
  end = offset + filesize - 1;
  if (window) {
  	pos = strchr(range, ',');
  	if (pos) {
  		*pos=0;
		start = strtoul(range,0,0);
		end = strtoul(pos+1,0,0);
	}
	else
		start = strtoul(range,0,0);
	if (start<offset) start = offset;
	if (end>(offset+filesize-1)) end = offset + filesize - 1;
  }
  
  if (end<offset) {
  	fprintf(stderr, "Disassembly window may not end before initial address.\n");
  	exit(-1);
  }

  if (start>(offset+filesize-1)) {
  	fprintf(stderr, "Disassembly window may not start beyond file size.\n");
  	exit(-1);
  }
  
  if (outfile) {
     ofp = fopen( outfile, "wb");
     if (!ofp) { fprintf(stderr, "Could not open file \"%s\" for writing.\n", outfile); exit(-1); }
  }

  seek_start = (start - offset) * 2;
  init_op_table();

  line = 1;
  address = start;

  if (disassembly) {

  /* pass 1a: build lists */

  if (fseek(ifp, seek_start, SEEK_SET) < 0)
   	 {
   	   fprintf(stderr, "Seek error on image file #%d at position %d\n", errno, 0);
   	   exit(-1);
   	 }

  while ((bytes_read = fread((void *)buffer, 1, rec_size, ifp)) > 0) {
        if (ferror(ifp)) {
   	      fprintf(stderr, "Read error on binary file (%d bytes read).\n", bytes_read);
   	      exit(-1);
   	}
  	for (i=0; (i < bytes_read) && (address <= end); i=i+2) {
  		cmd = buffer[i] * 256 + buffer[i+1];
  		check_label(address, cmd);

  		address++;
  	}
  }

  /* pass 1b: assign references to labels */
  
  address = start;

  if (fseek(ifp, seek_start, SEEK_SET) < 0)
   	 {
   	   fprintf(stderr, "Seek error on image file #%d at position %d\n", errno, 0);
   	   exit(-1);
   	 }

  while ((bytes_read = fread((void *)buffer, 1, rec_size, ifp)) > 0) {
        if (ferror(ifp)) {
   	      fprintf(stderr, "Read error on binary file (%d bytes read).\n", bytes_read);
   	      exit(-1);
   	}
  	for (i=0; (i < bytes_read) && (address <= end); i=i+2) {
  		cmd = buffer[i] * 256 + buffer[i+1];
  		label = search_label(internals, internals_size, address);

  		if (label) internals[label-1].value = cmd;
  		address++;
  	}
  }

  //for (i=0; i < internals_size; i++) printf("%04x: Lbl_%04i\n", internals[i].address, i+1);

  /* reposition */
  sort_label_list(internals, internals_size);
  sort_label_list(externals, externals_size);

  //for (i=0; i < internals_size; i++) printf("%04x: Lbl_%04i\n", internals[i].address, i+1);

  if (fseek(ifp, seek_start, SEEK_SET) < 0)
   	 {
   	   fprintf(stderr, "Seek error on image file #%d at position %d\n", errno, 0);
   	   exit(-1);
   	 }

  if (offset != 0) {
  	if (ascii) printf("     ");
  	if (octal) printf("                        ORG %06oB\n", offset);
  	else printf("                    ORG $%04x\n", offset);
  }

  /* output table of used base page symbols */

  if (list_basepage_symbols) {
  	Stringlist *sym_list = NULL, *indx, *node;

    if (numbers) printf("%3d ", line);
    printf("! Lower base page symbols (system constants)\n");
    line++;

  	for (i=0; i<512; i++) {
  		if (rom_labels_access[i] > 0) {
  			/* new list entry */
  			if (octal)
  				sprintf(str, " %s:%sEQU A+%06oB ! (%3i) %s", rom_labels[i], space + strlen(rom_labels[i]), i, rom_labels_access[i], rom_comments[i]);
  			else
  				sprintf(str, " %s:%sEQU A+$%04x ! (%3i) %s", rom_labels[i], space + strlen(rom_labels[i]), i, rom_labels_access[i], rom_comments[i]);
  			ordered_insert_string(&sym_list, str, 0);
	  	}
  	}

	for (indx=sym_list; indx; indx=(Stringlist *)indx->next) {
	    if (numbers) printf("%3d ", line);
		if (ascii) printf("     ");
		if (isource) printf ("ISOURCE");
		else {
			if (octal) printf("             ");
			else printf("         ");
		}
		printf("%s\n", indx->string);
		line++;
	}
  	
	for (indx=sym_list; indx;) {
		node = indx;
		indx = (Stringlist *)indx->next;
		free(node);
	}

    if (numbers) printf("%3d ", line);
    printf("! Upper base page symbols (system variables)\n");
    line++;

    sym_list = NULL;
  	for (i=0; i<512; i++) {
  		if (ram_labels_access[i] > 0) {
  			/* new list entry */
  			if (octal)
  				sprintf(str, " %s:%sEQU A+%06oB ! (%3i) %s", ram_labels[i], space + strlen(ram_labels[i]), i+0xfe00, ram_labels_access[i], ram_comments[i]);
  			else
  				sprintf(str, " %s:%sEQU A+$%04x ! (%3i) %s", ram_labels[i], space + strlen(ram_labels[i]), i+0xfe00, ram_labels_access[i], ram_comments[i]);
  			ordered_insert_string(&sym_list, str, 0);
	  	}
  	}

	for (indx=sym_list; indx; indx=(Stringlist *)indx->next) {
	    if (numbers) printf("%3d ", line);
		if (ascii) printf("     ");
		if (isource) printf ("ISOURCE");
		else {
			if (octal) printf("             ");
			else printf("         ");
		}
		printf("%s\n", indx->string);
		line++;
	}
  	
	for (indx=sym_list; indx;) {
		node = indx;
		indx = (Stringlist *)indx->next;
		free(node);
	}
	
    if (numbers) printf("%3d ", line);
	printf("! Program code\n");
	line++;
  }

  /* pass 2: produce disassembly */
  
  address = start;

  while ((bytes_read = fread((void *)buffer, 1, rec_size, ifp)) > 0) {
        if (ferror(ifp)) {
   	      fprintf(stderr, "Read error on binary file (%d bytes read).\n", bytes_read);
   	      exit(-1);
   	}
  	for (i=0; (i < bytes_read) && (address <= end); i=i+2) {
  		cmd = buffer[i] * 256 + buffer[i+1];
  		disassemble(line, address, cmd, ofp);
		line++;
  		address++;
  	}
  }
  

  /* output cross-reference tables */
  
  if (xref) {
      printf("\n\nCross Reference Table\n");
      printf("---------------------\n");
      printf("\nConstants:\n\n");
      print_xref(internals, internals_size, 1, 0);
      printf("\nVariables:\n\n");
      print_xref(internals, internals_size, 2, 0);
      printf("\nConstant vectors:\n\n");
      print_xref(internals, internals_size, 3, 0);
      printf("\nVariable vectors:\n\n");
      print_xref(internals, internals_size, 4, 0);
      printf("\nSubroutines:\n\n");
      print_xref(internals, internals_size, 5, 0);
      printf("\nJump points:\n\n");
      print_xref(internals, internals_size, 6, 0);
      printf("\nSubroutine vectors:\n\n");
      print_xref(internals, internals_size, 7, 0);
      printf("\nJump vectors:\n\n");
      print_xref(internals, internals_size, 8, 0);
      printf("\nBranches:\n\n");
      print_xref(internals, internals_size, 9, 0);
      printf("\nOthers:\n\n");
      print_xref(internals, internals_size, 0, 0);

      printf("\n\nExternal Reference Table\n");
      printf("------------------------\n");
      printf("\nConstants:\n\n");
      print_xref(externals, externals_size, 1, 1);
      printf("\nVariables:\n\n");
      print_xref(externals, externals_size, 2, 1);
      printf("\nConstant vectors:\n\n");
      print_xref(externals, externals_size, 3, 1);
      printf("\nVariable vectors:\n\n");
      print_xref(externals, externals_size, 4, 1);
      printf("\nSubroutines:\n\n");
      print_xref(externals, externals_size, 5, 1);
      printf("\nJump points:\n\n");
      print_xref(externals, externals_size, 6, 1);
      printf("\nSubroutine vectors:\n\n");
      print_xref(externals, externals_size, 7, 1);
      printf("\nJump vectors:\n\n");
      print_xref(externals, externals_size, 8, 1);
      printf("\nBranches:\n\n");
      print_xref(externals, externals_size, 9, 1);
      printf("\nOthers:\n\n");
      print_xref(externals, externals_size, 0, 1);
  }

  /* output hex/octal dump */
  
  if (dump) {
      printf ("\n\nMemory Dump\n");
      printf ("-----------\n\n");
      if (fseek(ifp, seek_start, SEEK_SET) < 0)
   	 {
   	   fprintf(stderr, "Seek error on image file #%d at position %d\n", errno, 0);
   	   exit(-1);
   	 }
   	 
	 address = start;

     while ((bytes_read = fread((void *)buffer, 1, rec_size, ifp)) > 0) {
        if (ferror(ifp)) {
   	      fprintf(stderr, "Read error on binary file (%d bytes read).\n", bytes_read);
   	      exit(-1);
   	}
  	for (i=0; (i < bytes_read) && (address <= end); i=i+16) {
  		if (octal) printf("%06o - ", address);
  		else printf("%04x - ", address);
  		for (j=0; j<8; j++) {
  		    cmd = buffer[i+j*2] * 256 + buffer[i+j*2+1];
  		    if (octal) printf("%06o ", cmd);
  		    else printf("%04x ", cmd);
  		}
  		for (j=0; j<16; j++) {
  		    c = buffer[i+j];
		    if ((c>31) && (c<127)) printf("%c", c);
		    else printf(".");
  		}
  		printf("\n");
  		address+=8;
  	}
      }
  }

  /* build string table */
  
  if (strings) {
  
      unsigned int str_start = 0;
      int min_len = 4;
      int index = 0;
      
      printf ("\n\nString Table\n");
      printf ("------------\n\n");
      if (fseek(ifp, seek_start, SEEK_SET) < 0)
   	 {
   	   fprintf(stderr, "Seek error on image file #%d at position %d\n", errno, 0);
   	   exit(-1);
   	 }
   	 
      address = start;

      while ((bytes_read = fread((void *)buffer, 1, rec_size, ifp)) > 0) {
        if (ferror(ifp)) {
   	      fprintf(stderr, "Read error on binary file (%d bytes read).\n", bytes_read);
   	      exit(-1);
   	}
  	for (i=0; (i < bytes_read) && (address <= end); i=i+2) {
  		for (j=0; j<2; j++) {
	  	    c = buffer[i+j];
	  	    if ((str_start) && ((c<32) || (c>126))) {			// Sequence end
	  		str[index] = 0;
	  		if (index >= min_len) {					// Threshold
		  	    if (octal) printf("%06o: \"%s\"\n", str_start, str);
		  	    else printf("%04x: \"%s\"\n", str_start, str);
		  	}
		  	str_start = 0;						// Reset
		  	index = 0;
	  	    }
	  	    else if ((c>31) && (c<127)) {		
	  		if ((j==0) && (!str_start)) str_start = address;	// New sequence
	  		if (str_start) str[index++] = c;
	  	    }
  		}
  		address++;
  	    }
     }

  }

  /* cleanup */
  
  for (i=0; i<internals_size; i++) {
      for (j=1; j<10; j++) {
          node = internals[i].ref[j];
          while (node) {
              tmp=node;
              node = node->next;
              free(tmp);
          }
      }
  }

  for (i=0; i<externals_size; i++) {
      for (j=1; j<10; j++) {
          node = externals[i].ref[j];
          while (node) {
              tmp=node;
              node = node->next;
              free(tmp);
          }
      }
  }
  
  } /* disassemble */
  
  /* ------------------------------------ assemble ---------------------------------- */
  else {
  psymbol = predefined;
  
  /* feed predefined symbols once */
  num_predefined = 0;
  for (;psymbol->name;psymbol++){
		tmp_symbol.name = psymbol->name;
		tmp_symbol.value = psymbol->value;
		tmp_symbol.type = ABS;
		tmp_symbol.vtype = ADDRESS;
		tmp_symbol.string = 0;
  		add_symbol(&tmp_symbol);
  		num_predefined++;
  }
  
  input_line = 1;
  listing_line = 1;

  /* assemble each module separately */
  while (assemble_module(ifp, ofp, &seek_start, &start, &input_line));
	
  } /* assemble */

  if (ifp) fclose(ifp);
  if (ofp) fclose(ofp);
  exit(0);
}
